<?xml version="1.0" encoding="utf-8"?>
<res class="ShaderProgram" class="ShaderProgram" path="Res://Metal/MetalBall.shader" Type="glsl" VertexShader="#version 450&#10;&#10;layout(binding = 0, std140) uniform UBO&#10;{&#10;    mat4 u_WorldMatrix;&#10;    mat4 u_ViewProjMatrix;&#10;} vs_ubo;&#10;&#10;layout(location = 0) in vec3 a_Position;&#10;layout(location = 0) out vec3 v_Position;&#10;layout(location = 1) out vec3 v_Normal;&#10;layout(location = 1) in vec3 a_Normal;&#10;&#10;void main()&#10;{&#10;    vec4 position = vs_ubo.u_WorldMatrix * vec4(a_Position, 1.0);&#10;    position = vs_ubo.u_ViewProjMatrix * position;&#10;    gl_Position = position;&#10;    v_Position = position.xyz;&#10;    v_Normal = normalize(vec3((vs_ubo.u_WorldMatrix * vec4(a_Normal, 0.0)).xyz));&#10;}&#10;&#10;" FragmentShader="#version 450&#10;&#10;struct PBRInfo&#10;{&#10;    float NdotL;&#10;    float NdotV;&#10;    float NdotH;&#10;    float LdotH;&#10;    float VdotH;&#10;    float perceptualRoughness;&#10;    float metalness;&#10;    vec3 reflectance0;&#10;    vec3 reflectance90;&#10;    float alphaRoughness;&#10;    vec3 diffuseColor;&#10;    vec3 specularColor;&#10;};&#10;&#10;layout(binding = 0, std140) uniform UBO&#10;{&#10;    vec3 u_CameraPosition;&#10;} fs_ubo;&#10;&#10;layout(location = 0) in vec3 v_Position;&#10;layout(location = 1) in vec3 v_Normal;&#10;layout(location = 0) out vec4 o_FragColor;&#10;&#10;vec3 SRgbToLinear(vec3 srgbIn)&#10;{&#10;    return pow(srgbIn, vec3(2.2000000476837158203125));&#10;}&#10;&#10;vec3 specularReflection(PBRInfo pbrInputs)&#10;{&#10;    return pbrInputs.reflectance0 + ((pbrInputs.reflectance90 - pbrInputs.reflectance0) * pow(clamp(1.0 - pbrInputs.VdotH, 0.0, 1.0), 5.0));&#10;}&#10;&#10;float geometricOcclusion(PBRInfo pbrInputs)&#10;{&#10;    float NdotL = pbrInputs.NdotL;&#10;    float NdotV = pbrInputs.NdotV;&#10;    float r = pbrInputs.alphaRoughness;&#10;    float attenuationL = (2.0 * NdotL) / (NdotL + sqrt((r * r) + ((1.0 - (r * r)) * (NdotL * NdotL))));&#10;    float attenuationV = (2.0 * NdotV) / (NdotV + sqrt((r * r) + ((1.0 - (r * r)) * (NdotV * NdotV))));&#10;    return attenuationL * attenuationV;&#10;}&#10;&#10;float microfacetDistribution(PBRInfo pbrInputs)&#10;{&#10;    float roughnessSq = pbrInputs.alphaRoughness * pbrInputs.alphaRoughness;&#10;    float f = (((pbrInputs.NdotH * roughnessSq) - pbrInputs.NdotH) * pbrInputs.NdotH) + 1.0;&#10;    return roughnessSq / ((3.1415927410125732421875 * f) * f);&#10;}&#10;&#10;vec3 diffuse(PBRInfo pbrInputs)&#10;{&#10;    return pbrInputs.diffuseColor / vec3(3.1415927410125732421875);&#10;}&#10;&#10;vec3 PbrLighting(vec3 pixelPosition, vec3 baseColor, vec3 normal, float metallic, float perceptualRoughness, vec3 eyePosition)&#10;{&#10;    float alphaRoughness = perceptualRoughness * perceptualRoughness;&#10;    vec3 f0 = vec3(0.039999999105930328369140625);&#10;    vec3 diffuseColor = baseColor * (vec3(1.0) - f0);&#10;    diffuseColor *= (1.0 - metallic);&#10;    vec3 specularColor = mix(f0, baseColor, vec3(metallic));&#10;    float reflectance = max(max(specularColor.x, specularColor.y), specularColor.z);&#10;    float reflectance90 = clamp(reflectance * 25.0, 0.0, 1.0);&#10;    vec3 specularEnvironmentR0 = specularColor;&#10;    vec3 specularEnvironmentR90 = vec3(1.0) * reflectance90;&#10;    vec3 n = normal;&#10;    vec3 v = normalize(eyePosition - pixelPosition);&#10;    float NdotV = abs(dot(n, v)) + 0.001000000047497451305389404296875;&#10;    vec3 _lightDir = vec3(0.57735025882720947265625);&#10;    vec3 param = vec3(1.2000000476837158203125);&#10;    vec3 _lightColor = SRgbToLinear(param);&#10;    vec3 l = normalize(_lightDir);&#10;    vec3 h = normalize(l + v);&#10;    vec3 reflection = -normalize(reflect(v, n));&#10;    float NdotL = clamp(dot(n, l), 0.001000000047497451305389404296875, 1.0);&#10;    float NdotH = clamp(dot(n, h), 0.0, 1.0);&#10;    float LdotH = clamp(dot(l, h), 0.0, 1.0);&#10;    float VdotH = clamp(dot(v, h), 0.0, 1.0);&#10;    PBRInfo pbrInputs = PBRInfo(NdotL, NdotV, NdotH, LdotH, VdotH, perceptualRoughness, metallic, specularEnvironmentR0, specularEnvironmentR90, alphaRoughness, diffuseColor, specularColor);&#10;    PBRInfo param_1 = pbrInputs;&#10;    vec3 F = specularReflection(param_1);&#10;    PBRInfo param_2 = pbrInputs;&#10;    float G = geometricOcclusion(param_2);&#10;    PBRInfo param_3 = pbrInputs;&#10;    float D = microfacetDistribution(param_3);&#10;    PBRInfo param_4 = pbrInputs;&#10;    vec3 diffuseContrib = (vec3(1.0) - F) * diffuse(param_4);&#10;    vec3 specContrib = ((F * G) * D) / vec3((4.0 * NdotL) * NdotV);&#10;    vec3 color = (_lightColor * NdotL) * (diffuseContrib + specContrib);&#10;    vec3 param_5 = vec3(0.300000011920928955078125);&#10;    vec3 _environmentLightColor = SRgbToLinear(param_5);&#10;    color += (baseColor * _environmentLightColor);&#10;    return color;&#10;}&#10;&#10;vec3 LinearToSRgb(vec3 linearIn)&#10;{&#10;    return pow(linearIn, vec3(0.4545454680919647216796875));&#10;}&#10;&#10;void main()&#10;{&#10;    float Float_24 = 0.300000011920928955078125;&#10;    vec4 Color_25_Value = vec4(1.0, 1.0, 0.000106999999843537807464599609375, 1.0);&#10;    float Float_28 = 1.0;&#10;    float Float_29 = 0.4799900054931640625;&#10;    vec3 Multiplication_27 = Color_25_Value.xyz * Float_28;&#10;    vec3 _BaseColor = Multiplication_27;&#10;    float _Metalic = Float_24;&#10;    float _PerceptualRoughness = Float_29;&#10;    float _Opacity = 1.0;&#10;    vec3 param = v_Position;&#10;    vec3 param_1 = _BaseColor;&#10;    vec3 param_2 = v_Normal;&#10;    float param_3 = _Metalic;&#10;    float param_4 = _PerceptualRoughness;&#10;    vec3 param_5 = fs_ubo.u_CameraPosition;&#10;    _BaseColor = PbrLighting(param, param_1, param_2, param_3, param_4, param_5);&#10;    vec3 param_6 = _BaseColor;&#10;    o_FragColor = vec4(LinearToSRgb(param_6), _Opacity);&#10;}&#10;&#10;" Graph="{&#10;    &quot;connections&quot;: [&#10;        {&#10;            &quot;in_id&quot;: &quot;{7783e48b-1039-4f9b-ae18-08ce08194b3c}&quot;,&#10;            &quot;in_index&quot;: 4,&#10;            &quot;out_id&quot;: &quot;{4e1aa5e3-82b7-4c12-b350-24eb7e1923b9}&quot;,&#10;            &quot;out_index&quot;: 0&#10;        },&#10;        {&#10;            &quot;in_id&quot;: &quot;{7783e48b-1039-4f9b-ae18-08ce08194b3c}&quot;,&#10;            &quot;in_index&quot;: 0,&#10;            &quot;out_id&quot;: &quot;{badfccbf-7a5d-4840-86a5-68b13c92f8e7}&quot;,&#10;            &quot;out_index&quot;: 0&#10;        },&#10;        {&#10;            &quot;converter&quot;: {&#10;                &quot;in&quot;: {&#10;                    &quot;id&quot;: &quot;any&quot;,&#10;                    &quot;name&quot;: &quot;B&quot;&#10;                },&#10;                &quot;out&quot;: {&#10;                    &quot;id&quot;: &quot;float&quot;,&#10;                    &quot;name&quot;: &quot;float&quot;&#10;                }&#10;            },&#10;            &quot;in_id&quot;: &quot;{badfccbf-7a5d-4840-86a5-68b13c92f8e7}&quot;,&#10;            &quot;in_index&quot;: 1,&#10;            &quot;out_id&quot;: &quot;{48703b22-79af-49f5-b341-63a2bf3e642b}&quot;,&#10;            &quot;out_index&quot;: 0&#10;        },&#10;        {&#10;            &quot;converter&quot;: {&#10;                &quot;in&quot;: {&#10;                    &quot;id&quot;: &quot;any&quot;,&#10;                    &quot;name&quot;: &quot;A&quot;&#10;                },&#10;                &quot;out&quot;: {&#10;                    &quot;id&quot;: &quot;vec3&quot;,&#10;                    &quot;name&quot;: &quot;rgb&quot;&#10;                }&#10;            },&#10;            &quot;in_id&quot;: &quot;{badfccbf-7a5d-4840-86a5-68b13c92f8e7}&quot;,&#10;            &quot;in_index&quot;: 0,&#10;            &quot;out_id&quot;: &quot;{70757413-9a6d-48de-bf5b-c8a30528c9ca}&quot;,&#10;            &quot;out_index&quot;: 0&#10;        },&#10;        {&#10;            &quot;in_id&quot;: &quot;{7783e48b-1039-4f9b-ae18-08ce08194b3c}&quot;,&#10;            &quot;in_index&quot;: 3,&#10;            &quot;out_id&quot;: &quot;{0ec53376-ccc5-4884-b305-8dcdad2327c6}&quot;,&#10;            &quot;out_index&quot;: 0&#10;        },&#10;        {&#10;            &quot;in_id&quot;: &quot;{7783e48b-1039-4f9b-ae18-08ce08194b3c}&quot;,&#10;            &quot;in_index&quot;: 2,&#10;            &quot;out_id&quot;: &quot;{41052098-fc71-4a3c-86bb-9176debfc2bb}&quot;,&#10;            &quot;out_index&quot;: 0&#10;        }&#10;    ],&#10;    &quot;nodes&quot;: [&#10;        {&#10;            &quot;id&quot;: &quot;{0ec53376-ccc5-4884-b305-8dcdad2327c6}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;name&quot;: &quot;Float&quot;,&#10;                &quot;number&quot;: &quot;0.3&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: -166,&#10;                &quot;y&quot;: 529&#10;            }&#10;        },&#10;        {&#10;            &quot;id&quot;: &quot;{7783e48b-1039-4f9b-ae18-08ce08194b3c}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;name&quot;: &quot;ShaderTemplate&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: 201,&#10;                &quot;y&quot;: 362&#10;            }&#10;        },&#10;        {&#10;            &quot;id&quot;: &quot;{70757413-9a6d-48de-bf5b-c8a30528c9ca}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;color&quot;: &quot;1 1 0.0156863 1 &quot;,&#10;                &quot;name&quot;: &quot;Color&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: -425,&#10;                &quot;y&quot;: 71&#10;            }&#10;        },&#10;        {&#10;            &quot;id&quot;: &quot;{48703b22-79af-49f5-b341-63a2bf3e642b}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;name&quot;: &quot;Float&quot;,&#10;                &quot;number&quot;: &quot;1.0&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: -357,&#10;                &quot;y&quot;: 295&#10;            }&#10;        },&#10;        {&#10;            &quot;id&quot;: &quot;{badfccbf-7a5d-4840-86a5-68b13c92f8e7}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;name&quot;: &quot;Multiplication&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: -145,&#10;                &quot;y&quot;: 227&#10;            }&#10;        },&#10;        {&#10;            &quot;id&quot;: &quot;{41052098-fc71-4a3c-86bb-9176debfc2bb}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;name&quot;: &quot;VertexAttribute&quot;,&#10;                &quot;option&quot;: &quot;normal&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: -239,&#10;                &quot;y&quot;: 435&#10;            }&#10;        },&#10;        {&#10;            &quot;id&quot;: &quot;{4e1aa5e3-82b7-4c12-b350-24eb7e1923b9}&quot;,&#10;            &quot;model&quot;: {&#10;                &quot;name&quot;: &quot;Float&quot;,&#10;                &quot;number&quot;: &quot;0.47999&quot;&#10;            },&#10;            &quot;position&quot;: {&#10;                &quot;x&quot;: -166,&#10;                &quot;y&quot;: 612&#10;            }&#10;        }&#10;    ]&#10;}&#10;" CullMode="CULL_BACK" BlendMode="Opaque" />
